@@ -69,9 +69,11 @@ class GradescopeClient:
         self.last_res = res = self.session.post(url, data=data, json=json, files=files, headers = headers)
         return res
 
-    def get_token(self, url, action=None, meta=None):
-        self.last_res = res = self.session.get(url)
-        self.last_soup = soup = BeautifulSoup(res.content, "html.parser")
+    def get_token(self, url, action=None, meta=None, content=None):
+        if not content:
+            self.last_res = res = self.session.get(url)
+            content = res.content
+        self.last_soup = soup = BeautifulSoup(content, "html.parser")
         form = None
         if action:
             form = soup.find("form", {"action":action})
@@ -1155,4 +1157,170 @@ class GradescopeClient:
         print()
         if found_repo:
             print("Found repo!")
-        return found_repo
\ No newline at end of file
+        return found_repo
+
+
+    def add_group_members(self, class_id: str, assignment_id: str, submission_id: str, user_ids: [str], check_group_size: bool=True):
+        """
+        Adds user IDs to a submission.
+        """
+        if not self.logged_in:
+            print("You must be logged in!")
+            return
+        refurl = self.base_url + f"/courses/{class_id}/assignments/{assignment_id}/submissions/{submission_id}"
+        url = refurl + f"/ownerships/many"
+        content = self.get_submission_data(class_id, assignment_id, submission_id)
+        if not content:
+            return
+        if check_group_size:
+            self.last_soup = soup = BeautifulSoup(content, "html.parser")
+            group_size = self.get_submission_members_limit(class_id, assignment_id, submission_id, submission_soup=soup)
+            if not group_size:
+                return
+            if not isinstance(user_ids, list):
+                user_ids = [user_ids]
+            if len(user_ids) > group_size - 1:
+                print("Attempted to add more group members than the assignment allows!")
+                return
+        token = self.get_token(refurl, meta="csrf-token", content=content)
+        payload = {
+            "user_ids[]": user_ids
+        }
+        self.last_res = res = self.submit_form(url, refurl, data=payload, header_token=token)
+        return res.ok
+
+    def remove_group_member(self, class_id: str, assignment_id: str, submission_id: str, emails: [str]):
+        """
+        Removes user IDs from a submission.
+        """
+        if not isinstance(emails, list):
+            emails = [emails]
+        if not self.logged_in:
+            print("You must be logged in!")
+            return
+        refurl = self.base_url + f"/courses/{class_id}/assignments/{assignment_id}/submissions/{submission_id}"
+        content = self.get_submission_data(class_id, assignment_id, submission_id)
+        if not content:
+            return
+        self.last_soup = soup = BeautifulSoup(content, "html.parser")
+        table = soup.find("tbody", class_="js-groupMemberList")
+        if not table:
+            return
+        students = table.find_all("tr", class_="")
+        stdmap = {} # Maps the student name to delete url
+        for student in students:
+            name = student.find("td", class_="table--cell-flex").text
+            url = self.base_url + student.find("form").get("action")
+            stdmap[name] = url
+
+        data = self.get_submission_members(class_id, assignment_id, submission_id, soup)
+
+        token = self.get_token(None, content=content, meta="csrf-token")
+        success = []
+        fdata = {
+            "authenticity_token": token,
+            "_method": "delete"
+        }
+        for email in emails:
+            for std in data:
+                if std.get("email") == email:
+                    del_url = stdmap.get(std.get("name"))
+                    if del_url:
+                        res = self.submit_form(del_url, refurl, data=fdata, header_token=token)
+                        success.append(res.ok)
+                    else:
+                        success.append(False)
+                    break
+            else:
+                success.append(False)
+        if len(emails) == 1:
+            return success[0]
+        return success
+        
+
+    def get_submission_members(self, class_id: str, assignment_id: str, submission_id: str, submission_soup=None) -> {str:str}:
+        """
+        Returns a dictionary of name and email of the students on the submission.
+        """
+        if not submission_soup:
+            content = self.get_submission_data(class_id, assignment_id, submission_id)
+            if not content:
+                return
+            self.last_soup = soup = BeautifulSoup(content, "html.parser")
+        else:
+            soup = submission_soup
+        cdata = GS_CDATA_decoder(soup=soup)
+        return cdata.get_gon().get("group_members")
+        
+
+    def get_submission_members_limit(self, class_id: str, assignment_id: str, submission_id: str, submission_soup=None) -> int:
+        """
+        Returns a dictionary of name and email of the students on the submission.
+        """
+        if not submission_soup:
+            content = self.get_submission_data(class_id, assignment_id, submission_id)
+            if not content:
+                return
+            self.last_soup = soup = BeautifulSoup(content, "html.parser")
+        else:
+            soup = submission_soup
+        cdata = GS_CDATA_decoder(soup=soup)
+        return cdata.get_gon().get("group_size")
+
+
+    def get_groups_roster(self, class_id: str, assignment_id: str, submission_id: str, submission_soup=None):
+        """
+        Gets all of the roster data of all students who can be put on a submission.
+        """
+        if submission_soup is None:
+            content = self.get_submission_data(class_id, assignment_id, submission_id)
+            if not content:
+                return
+            self.last_soup = soup = BeautifulSoup(content, "html.parser")
+        else:
+            soup = submission_soup
+        cdata = GS_CDATA_decoder(soup=soup)
+        return cdata.get_gon().get("roster")
+
+    def get_groups_data_from_emails(self, class_id: str, assignment_id: str, submission_id: str, emails: [str]):
+        roster = self.get_groups_roster(class_id, assignment_id, submission_id)
+        if roster:
+            if isinstance(emails, str):
+                emails = [emails]
+            ids = []
+            
+            for email in emails:
+                for student in roster:
+                    if student.get("email") == email:
+                        ids.append(student)
+                        break
+                else:
+                    ids.append(False)
+
+            if len(emails) == 1:
+                return ids[0]
+            return ids
+
+    def duplicate_assignment(self, class_id: str, assignment_id: str, title: str):
+        """
+        class_id is the class you want to duplicate the assignment into.
+        assignment_id is the assignment you want to duplicate into that class.
+        title is the title of the assignment.
+        """
+        if not self.logged_in:
+            print("You must be logged in to duplicate an assignment!")
+            return
+        base_url = f"https://www.gradescope.com"
+        location_url = f"/courses/{class_id}/assignments"
+        referer_url = base_url + location_url + "/duplicate"
+        url = base_url + location_url
+        token = self.get_token(referer_url, meta="csrf-token")
+        payload = {
+                "utf8": "✓",
+                "authenticity_token": token,
+                "assignment[id]": assignment_id,
+                "assignment[title]": title,
+            }
+        self.last_res = res = self.submit_form(url, referer_url, data=payload)
+        import ipdb; ipdb.set_trace()
+        return res.ok
